@@ -1,6 +1,6 @@
 #!/usr/bin/perl -w
 
-# $Id: ClamdOmitScan.pl,v 1.10 2007/08/19 19:10:41 okamura Exp $
+# $Id: ClamdOmitScan.pl,v 1.15 2007/09/23 03:05:10 okamura Exp $
 
 require 5.8.0;
 use strict;
@@ -20,6 +20,9 @@ use File::Path;
 use Fcntl	qw(:flock);
 use IO::Socket;
 use Time::Local;
+eval {
+	require Sys::Filesystem;
+};
 
 # プロジェクト
 
@@ -46,7 +49,7 @@ use vars qw(
 $ProgramName = 'ClamdOmitScan';
 
 # バージョン
-$Version = 'ClamdOmitScan.pl v1.0.1';
+$Version = 'ClamdOmitScan.pl v1.1.1';
 
 # プレフィクス
 $Prefix = '/usr/local';
@@ -76,6 +79,7 @@ $Prefix = '/usr/local';
 	UknownOpts		=> 40,
 	BadParams		=> 41,
 	BadArguments	=> 42,
+	UnsupportedOpts	=> 43,
 	ScanError		=> 50,
 	MaintainError	=> 51,
 );
@@ -165,17 +169,26 @@ sub WarnUnsupportedOption {
 		Warn 'Ignoring option %s', $name;
 	}
 	if ($Conf{'error-unsupported'}) {
-		exit $ExitStatus{UnknownOpts};
+		exit $ExitStatus{UnsupportedOpts};
 	}
 }
 
+# ファイル名のエスケープ
+# @param $path	ファイル名
+sub EscapedPath {
+	my	($path) = @_;
+
+	$path =~ s/([\x00-\x1F\x7f])/?/gs;
+	return $path;
+}
+
 # 除外レポート
 # @param $path パス
 sub ReportExcluded {
 	unless ($Conf{infected}) {
 		my	($path) = @_;
 
-		Output "%s: Excluded\n", $path;
+		Output "%s: Excluded\n", EscapedPath($path);
 	}
 }
 
@@ -184,7 +197,7 @@ sub ReportExcluded {
 sub ReportChecked {
 	my	($path) = @_;
 
-	Output "%s: Checked\n", $path	unless ($Conf{infected});
+	Output "%s: Checked\n", EscapedPath($path)	unless ($Conf{infected});
 	$Summary{count}->{files}->{checked}++;
 	$Summary{count}->{data}->{checked} += $ActionSwitch{statCache}->[7];
 }
@@ -214,13 +227,13 @@ sub ReportResult {
 	}
 
 	foreach (@{$result{UNKNOWN}}) {
-		ReportUnknownResult($path, $_);
+		ReportUnknownResult($_->[0], $_->[1]);
 	}
 
 	if (scalar(@{$result{FOUND}})) {
 		if ($Conf{remove}) {
 			unless (unlink $path) {
-				Warn 'Can\'t unlink: %s', $path;
+				Warn 'Can\'t unlink: %s', EscapPath($path);
 			}
 		}
 		else {
@@ -234,7 +247,7 @@ sub ReportResult {
 sub ReportOkResult {
 	my	($path) = @_;
 
-	Output "%s: OK\n", $path	unless ($Conf{infected});
+	Output "%s: OK\n", EscapedPath($path)	unless ($Conf{infected});
 	$Summary{count}->{files}->{scaned}++;
 	$Summary{count}->{data}->{scaned}
 		+= $ActionSwitch{statCache}->[7]
@@ -248,7 +261,7 @@ sub ReportOkResult {
 sub ReportFoundResult {
 	my	($path, $virus) = @_;
 
-	Output "%s: %s FOUND\n", $path, $virus;
+	Output "%s: %s FOUND\n", EscapedPath($path), $virus;
 	$Summary{count}->{files}->{scaned}++;
 	$Summary{count}->{data}->{scaned}
 		+= $ActionSwitch{statCache}->[7]
@@ -264,7 +277,7 @@ sub ReportFoundResult {
 sub ReportErrorResult {
 	my	($path, $message) = @_;
 
-	Output "%s: %s ERROR\n", $path, $message;
+	Output "%s: %s ERROR\n", EscapedPath($path), $message;
 	$Summary{count}->{error}++;
 }
 
@@ -274,7 +287,7 @@ sub ReportErrorResult {
 sub ReportUnknownResult {
 	my	($path, $message) = @_;
 
-	Output "%s: %s\n", $path, $message;
+	Output "%s: %s\n", EscapedPath($path), $message;
 	$Summary{count}->{error}++;
 }
 
@@ -337,7 +350,7 @@ sub OpenScanData {
 	# 絶対に root ディレクトリを使用しては駄目!
 	if ($handle->{basePath} eq File::Spec->rootdir()) {
 		Error '%s is not allow as scanning data directory.',
-			$handle->{basePath}
+			EscapPath($handle->{basePath})
 		;
 		return undef;
 	}
@@ -350,10 +363,13 @@ sub OpenScanData {
 			@created = mkpath($handle->{basePath}, 0,  $Conf{'perm-root'})
 		};
 		if ($@) {
-			Error 'Can\'t create directory: %s: %s', $@, $handle->{basePath};
+			Error 'Can\'t create directory: %s: %s',
+				$@, EscapedPath($handle->{basePath})
+			;
 			return undef;
 		}
 		chmod $Conf{'perm-root'}, $handle->{basePath};
+		@created = map { EscapedPath($_) } @created;
 		Debug "Create directories: %s", "@created";
 	}
 
@@ -368,12 +384,14 @@ sub OpenScanData {
 	$ret = open $handle->{fileHandle}, '>', $lockFile;
 	umask $oldUmask;
 	unless ($ret) {
-		Error '%s: %s', "$!", $lockFile;
+		Error '%s: %s', "$!", EscapedPath($lockFile);
 		return undef;
 	}
 	chmod $Conf{'perm-lock'}, $lockFile;
 
-	Debug "Lock (%s) %s ...", $lock ? 'exclusive' : 'shared', $lockFile;
+	Debug "Lock (%s) %s ...",
+		$lock ? 'exclusive' : 'shared', EscapedPath($lockFile)
+	;
 	flock $handle->{fileHandle}, $lock ? LOCK_EX : LOCK_SH;
 	Debug "Done.";
 
@@ -397,19 +415,19 @@ sub CloseScanData {
 
 	unless (flock $handle->{fileHandle}, LOCK_UN) {
 		$ret = 0;
-		Warn 'Can\'t unlock a file: %s: %s', "$!", $lockFile;
+		Warn 'Can\'t unlock a file: %s: %s', "$!", EscapedPath($lockFile);
 	}
-	Debug '%s is unlocked.', $lockFile;
+	Debug '%s is unlocked.', EscapedPath($lockFile);
 
 	unless (close $handle->{fileHandle}) {
 		$ret = 0;
-		Warn 'Can\'t close a file: %s: %s', "$!", $lockFile;
+		Warn 'Can\'t close a file: %s: %s', "$!", EscapedPath($lockFile);
 	}
 
 	# ロックファイルは削除しない
 #	unless (unlink $lockFile) {
 #		$ret = 0;
-#		Warn 'Can\'t unlink a file: %s: %s', "$!", $lockFile;
+#		Warn 'Can\'t unlink a file: %s: %s', "$!", EscapedPath($lockFile);
 #	}
 
 	return $ret;
@@ -424,7 +442,7 @@ sub _SlimScanData {
 		if (-d $dataPath) {
 			unless (-d $realPath) {
 				unless (rmtree($dataPath, 0, 0)) {
-					Warn '%s: %s', "$!", $dataPath;
+					Warn '%s: %s', "$!", EscapedPath($dataPath);
 				}
 				$File::Find::prune = 1;
 			}
@@ -432,14 +450,14 @@ sub _SlimScanData {
 		elsif (-f $dataPath) {
 			unless (-f $realPath) {
 				unless (unlink($dataPath)) {
-					Warn '%s: %s', "$!", $dataPath;
+					Warn '%s: %s', "$!", EscapedPath($dataPath);
 				}
 			}
 		}
 		else {
-			Warn 'This is not a file or directory: %s', $dataPath;
+			Warn 'This is not a file or directory: %s', EscapedPath($dataPath);
 			unless (unlink($dataPath)) {
-				Warn '%s: %s', "$!", $dataPath;
+				Warn '%s: %s', "$!", EscapedPath($dataPath);
 			}
 		}
 	}
@@ -457,7 +475,7 @@ sub _SlimScanData4Maintain {
 			}
 			else {
 				unless (rmtree($dataPath, 0, 0)) {
-					Warn '%s: %s', "$!", $dataPath;
+					Warn '%s: %s', "$!", EscapedPath($dataPath);
 				}
 				$File::Find::prune = 1;
 			}
@@ -472,21 +490,21 @@ sub _SlimScanData4Maintain {
 				else {
 					if (-e $dataPath) {	# _LastScanClamavVersion で消されるかも
 						unless (unlink $dataPath) {
-							Warn '%s: %s', "$!", $dataPath;
+							Warn '%s: %s', "$!", EscapedPath($dataPath);
 						}
 					}
 				}
 			}
 			else {
 				unless (unlink($dataPath)) {
-					Warn '%s: %s', "$!", $dataPath;
+					Warn '%s: %s', "$!", EscapedPath($dataPath);
 				}
 			}
 		}
 		else {
-			Warn 'This is not a file or directory: %s', $dataPath;
+			Warn 'This is not a file or directory: %s', EscapedPath($dataPath);
 			unless (unlink($dataPath)) {
-				Warn '%s: %s', "$!", $dataPath;
+				Warn '%s: %s', "$!", EscapedPath($dataPath);
 			}
 		}
 	}
@@ -558,15 +576,15 @@ sub _LastScanClamavVersion {
 	my	$line;
 
 	unless (open $fh, '<', $path) {
-		Warn '%s: %s', "$!", $path;
+		Warn '%s: %s', "$!", EscapedPath($path);
 		if (-f $path) {
 			unless (unlink $path) {
-				Warn '%s: %s', "$!", $path;
+				Warn '%s: %s', "$!", EscapedPath($path);
 			}
 		}
 		elsif (-d $path) {
 			unless (rmtree($path, 0, 0)) {
-				Warn '%s: %s', "$!", $path;
+				Warn '%s: %s', "$!", EscapedPath($path);
 			}
 		}
 		return %result;
@@ -577,15 +595,15 @@ sub _LastScanClamavVersion {
 
 	%result = ParseVersion($line);
 	unless (defined $result{ClamAV}) {
-		Warn 'Invalid version format: %s: %s', $path, $line;
+		Warn 'Invalid version format: %s: %s', EscapedPath($path), $line;
 		if (-f $path) {
 			unless (unlink $path) {
-				Warn '%s: %s', "$!", $path;
+				Warn '%s: %s', "$!", EscapedPath($path);
 			}
 		}
 		elsif (-d $path) {
 			unless (rmtree($path, 0, 0)) {
-				Warn '%s: %s', "$!", $path;
+				Warn '%s: %s', "$!", EscapedPath($path);
 			}
 		}
 	}
@@ -639,7 +657,7 @@ sub AddScanData {
 		my	@created =  mkpath([$dir], 0, $Conf{'perm-dir'});
 
 		unless (0 < scalar(@created)) {
-			Warn 'Can\'t create a directory: %s: %s', "$!", $dir;
+			Warn 'Can\'t create a directory: %s: %s', "$!", EscapedPath($dir);
 			$ret = 0;
 		}
 		foreach (@created) {
@@ -656,12 +674,12 @@ sub AddScanData {
 			chmod $Conf{'perm-file'}, $scanData;
 		}
 		else {
-			Warn '%s: %s', "$!", $scanData;
+			Warn '%s: %s', "$!", EscapedPath($scanData);
 			$ret = 0;
 		}
 	}
 	else {
-		Warn 'Is not a file: %s', $path;
+		Warn 'Is not a file: %s', EscapedPath($path);
 		$ret = 0;
 	}
 
@@ -694,13 +712,13 @@ sub UpdateScanData {
 				close $fh;
 			}
 			else {
-				Warn '%s: %s', "$!", $scanData;
+				Warn '%s: %s', "$!", EscapedPath($scanData);
 				$ret = 0;
 			}
 			umask $oldUmask;
 		}
 		else {
-			Warn 'Is not a file: %s', $scanData;
+			Warn 'Is not a file: %s', EscapedPath($scanData);
 			$ret = 0;
 		}
 	}
@@ -750,6 +768,26 @@ sub Initialize {
 		return 0;
 	}
 
+	if ($Conf{rsrc}) {
+		my	$fs = Sys::Filesystem->new();
+
+		$ActionSwitch{filesystem} = [
+			map {
+				{
+					mountPoint	=> $fs->mount_point($_),
+					type		=> $fs->type($_),
+				};
+			} sort {
+				-1 * (
+					length($fs->mount_point($a))
+						<=>
+					length($fs->mount_point($b))
+				)
+			} $fs->mounted_filesystems()
+		];
+		Debug 'Got filesystem informations.';
+	}
+
 	unless (
 		$ActionSwitch{ScanDataHandle} = OpenScanData($Conf{data}, $Conf{lock})
 	) {
@@ -783,7 +821,7 @@ sub InitializeOutputStream {
 			push @{$ActionSwitch{outputStream}}, $fh;
 		}
 		else {
-			Error "%s: %s", "$!", $Conf{log};
+			Error "%s: %s", "$!", EscapedPath($Conf{log});
 			return 0;
 		}
 	}
@@ -894,7 +932,7 @@ sub InitializeWithClamdConf {
 		}
 	}
 	elsif (defined $Conf{'config-file'}) {
-		Warn '%s: %s', "$!", $Conf{'config-file'};
+		Warn '%s: %s', "$!", EscapedPath($Conf{'config-file'});
 	}
 
 	# デフォルト値を設定する
@@ -915,10 +953,13 @@ sub InitializeScanType {
 	# UNIX ドメインソケット
 	if (defined($Conf{'socket'}) or !defined($Conf{port})) {
 		unless (defined $Conf{'socket'}) {
-			$Conf{'socket'} = '/tmp/clamd';
+			$Conf{'socket'} = '/tmp/clamd.socket';
 		}
 		$ActionSwitch{scanType} = 'unix';
-		$ActionSwitch{ScanFunc} = \&ScanFile;
+		$ActionSwitch{ScanFunc} = $Conf{rsrc}
+			? \&ScanFileWithRsrc
+			: \&ScanFile
+		;
 		if ($Conf{multiscan}) {
 			$ActionSwitch{ClamdScanCommand} = 'MULTISCAN';
 		}
@@ -930,14 +971,16 @@ sub InitializeScanType {
 		}
 		# UNIX ドメインソケットを使用しても STREAM をするときはデータ送信に
 		# INET ソケットを使用しなければならない。だから host を設定する。
-		# UNIX ドメインソケットはローカルホストのみだから 127.0.0.1。
-		$Conf{host} = '127.0.0.1';
+		$Conf{host} = '127.0.0.1'	unless (defined $Conf{host});
 	}
 	# INET ソケット
 	else {
 		$Conf{host} = '127.0.0.1'	unless (defined $Conf{host});
 		$ActionSwitch{scanType} = 'inet';
-		$ActionSwitch{ScanFunc} = \&ScanStream;
+		$ActionSwitch{ScanFunc} = $Conf{rsrc}
+			? \&ScanStreamWithRsrc
+			: \&ScanStream
+		;
 	}
 
 	return 1;
@@ -951,15 +994,15 @@ sub SetMoveDir {
 	my	($name, $value) = @_;
 
 	unless (-e $value) {
-		Error 'No such directory: %s', $value;
+		Error 'No such directory: %s', EscapedPath($value);
 		return 0;
 	}
 	unless (-d $value) {
-		Error 'Not a directory: %s', $value;
+		Error 'Not a directory: %s', EscapedPath($value);
 		return 0;
 	}
 	unless (-w $value and -x $value) {
-		Error 'Permission denied: %s', $value;
+		Error 'Permission denied: %s', EscapedPath($value);
 		return 0;
 	}
 
@@ -967,6 +1010,21 @@ sub SetMoveDir {
 	return 1;
 }
 
+# HFS 判定
+# @param $path	パス
+# @return HFS ファイルシステム上のパスかどうか
+sub IsHfsPath {
+	my	($path) = @_;
+
+	foreach my $fs (@{$ActionSwitch{filesystem}}) {
+		if (substr($path, 0, length($fs->{mountPoint})) eq $fs->{mountPoint}) {
+			return $fs->{type} eq 'hfs'; 
+		}
+	}
+
+	return 0;
+}
+
 #--- 終了処理 ----
 # 終了処理
 sub Terminate {
@@ -1093,6 +1151,9 @@ sub DoesNeedScan {
 		my	@stat =  stat $path;
 
 		$ActionSwitch{statCache} = [@stat];
+		if ($Conf{rsrc} and $ActionSwitch{onHfsDir}) {
+			$ActionSwitch{statCache}->[7] += (stat "$path/rsrc")[7];
+		}
 		$mtime = $stat[9] <= $stat[10] ? $stat[10] : $stat[9];
 	}
 
@@ -1120,7 +1181,7 @@ sub DoesNeedScan {
 				and
 			$version{db} eq $ActionSwitch{ClamavVersion}->{hash}->{db}
 		) {
-			Info 'Already scaned with current version: %s', $path;
+			Info 'Already scaned with current version: %s', EscapedPath($path);
 			return 0;
 		}
 	}
@@ -1132,7 +1193,7 @@ sub DoesNeedScan {
 			<=
 		$ActionSwitch{ClamavVersion}->{hash}->{dbBuildDateTime}
 	) {
-		Info 'Inspection period is over: %s', $path;
+		Info 'Inspection period is over: %s', EscapedPath($path);
 		return 0;
 	}
 
@@ -1204,7 +1265,7 @@ sub Scan {
 	my	($target) = @_;
 
 	$target = realpath($target);
-	Debug 'Scan target: %s', $target;
+	Debug 'Scan target: %s', EscapedPath($target);
 
 	unless (SlimScanData($ActionSwitch{ScanDataHandle}, $target)) {
 		return 0;
@@ -1213,6 +1274,10 @@ sub Scan {
 	if (-f $target) {
 		my	$dir = dirname($target);
 
+		if ($Conf{rsrc}) {
+			$ActionSwitch{onHfsDir} = IsHfsPath($target);
+		}
+
 		if (!IsScanTargetDirectory($dir)) {
 			ReportExcluded($target);
 			$target = undef;
@@ -1225,12 +1290,22 @@ sub Scan {
 			ReportChecked($target);
 			$target = undef;
 		}
-		elsif (-z $target) {
+		elsif (
+			-z $target
+				and
+			(
+				!$Conf{rsrc}
+					or
+				!IsHfsPath($target)
+					or
+				-z "$target/rsrc"
+			)
+		) {
 			ReportOkResult($target);
 		}
 	}
 	elsif (!-d $target) {
-		Warn "Is not a file or directory: %s", $target;
+		Warn "Is not a file or directory: %s", EscapedPath($target);
 		$target = undef;
 	}
 	elsif (!$Conf{recursive}) {
@@ -1242,12 +1317,20 @@ sub Scan {
 		find(
 			{
 				wanted		=> \&_Scan,
-				preprocess	=> sub {
-					my	@list = @_;
+				preprocess	=> $Conf{rsrc}
+					? sub {
+						my	@list = @_;
 
-					$ActionSwitch{dirDepth}++;
-					return @list;
-				},
+						$ActionSwitch{onHfsDir} = IsHfsPath($File::Find::dir);
+						$ActionSwitch{dirDepth}++;
+						return @list;
+					}
+					: sub {
+						my	@list = @_;
+
+						$ActionSwitch{dirDepth}++;
+						return @list;
+					},
 				postprocess => sub {
 					$ActionSwitch{dirDepth}--;
 				},
@@ -1268,7 +1351,7 @@ sub _Scan {
 			$Conf{'max-dir-recursion'} < $ActionSwitch{dirDepth}
 		) {
 			$File::Find::prune = 1;
-			Debug '%s: Excluded', $path;
+			Debug '%s: Excluded', EscapedPath($path);
 		}
 		else {
 			$Summary{count}->{directories}++;
@@ -1281,14 +1364,24 @@ sub _Scan {
 		elsif (!DoesNeedScan($path)) {
 			ReportChecked($path);
 		}
-		elsif (-z $path) {
+		elsif (
+			-z $path
+				and
+			(
+				!$Conf{rsrc}
+					or
+				!$ActionSwitch{onHfsDir}
+					or
+				-z "$path/rsrc"
+			)
+		) {
 			ReportOkResult($path);
 		}
 		else {
 			my	$timeout = GetScanTimeout($path);
 			my	%result;
 
-			Debug 'timeout: %.3f s: %s', $timeout, $path;
+			Debug 'timeout: %.3f s: %s', $timeout, EscapedPath($path);
 			eval {
 				local	$SIG{ALRM};
 
@@ -1302,13 +1395,13 @@ sub _Scan {
 				chomp $@;
 				if ($@ =~ m/^clamd timeout/) {
 					$ActionSwitch{timeoutFreq}++;
-					Warn 'Scan timeout: %s', $path;
+					Warn 'Scan timeout: %s', EscapedPath($path);
 					if ($Conf{'timeout-command'}) {
 						system $Conf{'timeout-command'};
 					}
 				}
 				else {
-					Warn '%s: %s', $@, $path;
+					Warn '%s: %s', $@, EscapedPath($path);
 				}
 
 				# 少し待ってからクローズする。
@@ -1398,6 +1491,56 @@ sub ScanFile {
 	return %result;
 }
 
+# リソースフォークを含むファイルスキャン
+# @param $path ファイルのパス
+# @return 次の形式のハッシュ
+#	(
+#		OK	=> [ファイル名, ...],
+#		FOUND	=> [[ファイル名, ウィルス名], ...],
+#		ERROR	=> [[ファイル名, エラーメッセージ], ...],
+#		UNKNOWN	=> [[ファイル名, メッセージ], ...],
+#	)
+sub ScanFileWithRsrc {
+	unless ($ActionSwitch{onHfsDir}) {
+		return ScanFile(@_);
+	}
+
+	my	($path) = @_;
+	my	%result = (OK => [], FOUND => [], ERROR => [], UNKNOWN => []);
+
+	if (-z $path) {
+		push @{$result{OK}}, $path;
+	}
+	else {
+		my	%forkResult = ScanFile($path);
+
+		foreach my $key (keys %result) {
+			push @{$result{$key}}, @{$forkResult{$key}};
+		}
+	}
+
+	if (-z "$path/rsrc") {
+		push @{$result{OK}}, $path;
+	}
+	else {
+		my	%forkResult = ScanFile("$path/rsrc");
+
+		ReplacePathInResult($path, \%forkResult);
+		foreach my $key (keys %result) {
+			push @{$result{$key}}, @{$forkResult{$key}};
+		}
+	}
+
+	if (scalar(@{$result{OK}}) == 2) {
+		pop @{$result{OK}};
+	}
+	else {
+		$result{OK} = [];
+	}
+
+	return %result;
+}
+
 # ファイルのストリームスキャン
 # @param $path ファイルのパス
 # @return 次の形式のハッシュ
@@ -1431,18 +1574,68 @@ sub ScanStream {
 				ReplacePathInResult($path, \%result);
 			}
 			else {
-				Warn '%s: %s', "$!", $path;
+				Warn '%s: %s', "$!", EscapedPath($path);
 			}
 			CloseClamd($conn);
 		}
 		else {
-			Error '%s: %s', "$!", $path;
+			Error '%s: %s', "$!", EscapedPath($path);
 		}
 	}
 
 	return %result;
 }
 
+# リソースフォークを含むファイルスキャン
+# @param $path ファイルのパス
+# @return 次の形式のハッシュ
+#	(
+#		OK	=> [ファイル名, ...],
+#		FOUND	=> [[ファイル名, ウィルス名], ...],
+#		ERROR	=> [[ファイル名, エラーメッセージ], ...],
+#		UNKNOWN	=> [[ファイル名, メッセージ], ...],
+#	)
+sub ScanStreamWithRsrc {
+	unless ($ActionSwitch{onHfsDir}) {
+		return ScanStream(@_);
+	}
+
+	my	($path) = @_;
+	my	%result = (OK => [], FOUND => [], ERROR => [], UNKNOWN => []);
+
+	if (-z $path) {
+		push @{$result{OK}}, $path;
+	}
+	else {
+		my	%forkResult = ScanStream($path);
+
+		foreach my $key (keys %result) {
+			push @{$result{$key}}, @{$forkResult{$key}};
+		}
+	}
+
+	if (-z "$path/rsrc") {
+		push @{$result{OK}}, $path;
+	}
+	else {
+		my	%forkResult = ScanStream("$path/rsrc");
+
+		ReplacePathInResult($path, \%forkResult);
+		foreach my $key (keys %result) {
+			push @{$result{$key}}, @{$forkResult{$key}};
+		}
+	}
+
+	if (scalar(@{$result{OK}}) == 2) {
+		pop @{$result{OK}};
+	}
+	else {
+		$result{OK} = [];
+	}
+
+	return %result;
+}
+
 # スキャン結果のファイルパス入れ替え
 # @param $path パス名
 # @param $result スキャン結果のハッシュへのリファイレンス
@@ -1555,7 +1748,7 @@ sub SendClamdStreamCommand {
 			while ($bufsiz = read($fh, $buffer, $Conf{bufsiz})) {
 				unless (SendData2Socket($dconn, $buffer, $bufsiz)) {
 					push @{$result{ERROR}},
-						[undef, "Can't connet send data to data port $port."]
+						[undef, "Can't send data to data port $port."]
 					;
 					last;
 				}
@@ -1698,7 +1891,7 @@ sub ClamavVersion {
 	}
     else {
 		if ($ActionSwitch{scanType} eq 'unix') {
-			Error '%s: %s', "$!", $Conf{socket};
+			Error '%s: %s', "$!", EscapedPath($Conf{socket});
 		}
 		else {
 			Error '%s: %s:%d', "$!", $Conf{host}, $Conf{port};
@@ -1790,7 +1983,8 @@ sub MoveInfected {
 		# 後からパーミッションが変えられるかもしれないからチェックする。
 		unless (-w $moveDir and -x $moveDir) {
 			Warn "Can't %s file '%s': cannot write to '%s'",
-				defined($Conf{move}) ? 'move' : 'copy', $path, $moveDir
+				defined($Conf{move}) ? 'move' : 'copy',
+				EscapedPath($path), EscapedPath($moveDir)
 			;
 			return undef;
 		}
@@ -1803,7 +1997,7 @@ sub MoveInfected {
 		my	$newMovedPath = $movedPath;
 
 		unless ((stat($movedPath))[1] != $ActionSwitch{statCache}->[1]) {
-			Warn "File excluded '%s'", $path;
+			Warn "File excluded '%s'", EscapedPath($path);
 			return 1;
 		}
 		for (my $i = 0; $i < 1000; $i++) {
@@ -1820,7 +2014,7 @@ sub MoveInfected {
 		if (!CopyFile($path, $movedPath)) {
 			Warn "Can't %s '%s' to '%s': %s",
 				defined($Conf{move}) ? 'move' : 'copy',
-				$path, $movedPath, "$!"
+				EscapedPath($path), EscapedPath($movedPath), "$!"
 			;
 			# 移動先が既存のファイルでもどうせ書き込んでしまっているから
 			# 消してしまえ! それもできなきゃ諦めろ。
@@ -1837,7 +2031,7 @@ sub MoveInfected {
 	# 必要ならば元ファイル削除
 	if (defined($Conf{move}) and -f $path) {
 		unless (unlink $path) {
-			Warn "Can't unlink '%s': %s", $path, "$!";
+			Warn "Can't unlink '%s': %s", EscapedPath($path), "$!";
 			return undef;
 		}
 	}
@@ -1854,31 +2048,31 @@ sub CopyFile {
 	my	($src, $dst) = @_;
 
 	# Mac OS X の場合
-	if (-f '/usr/bin/ditto' and -x '/usr/bin/ditto') {
+	if ($^O eq 'darwin' and -f '/usr/bin/ditto' and -x '/usr/bin/ditto') {
 		# リソースフォークを維持するために ditto コマンドを使用する。
 		system('/usr/bin/ditto', '--rsrc', $src, $dst);
 		if ($? == -1) {
 			Warn "%s: /usr/bin/ditto --rsrc '%s' '%s'",
-				"$!", $src, $dst
+				"$!", EscapedPath($src), EscapedPath($dst)
 			;
 			return 0;
 		}
 		elsif ($? & 127) {
 			if ($? & 128) {
 				Warn "Died with coredump: %d: /usr/bin/ditto --rsrc '%s' '%s'",
-					($? & 127), $src, $dst
+					($? & 127), EscapedPath($src), EscapedPath($dst)
 				;
 			}
 			else {
 				Warn "Died: %d: /usr/bin/ditto --rsrc '%s' '%s'",
-					($? & 127), $src, $dst
+					($? & 127), EscapedPath($src), EscapedPath($dst)
 				;
 			}
 			return 0;
 		}
 		elsif ($?) {
 			Warn "child exited: %d: /usr/bin/ditto --rsrc '%s' '%s'",
-				$? >> 8, $src, $dst
+				$? >> 8, EscapedPath($src), EscapedPath($dst)
 			;
 			return 0;
 		}
@@ -1890,12 +2084,12 @@ sub CopyFile {
 		my	$redLen;
 
 		unless (open $sfh, '<', $src) {
-			Warn "%s: %s", "$!", $src;
+			Warn "%s: %s", "$!", EscapedPath($src);
 			return 0;
 		}
 		unless (open $dfh, '>', $dst) {
 			close $sfh;
-			Warn "%s: %s", "$!", $src;
+			Warn "%s: %s", "$!", EscapedPath($src);
 			return 0;
 		}
 
@@ -1903,14 +2097,14 @@ sub CopyFile {
 			my	$wroteLen = syswrite $dfh, $buffer, $redLen;
 
 			unless ($wroteLen == $redLen) {
-				Warn "%s: %s", "$!", $src;
+				Warn "%s: %s", "$!", EscapedPath($src);
 				close $sfh;
 				close $dfh;
 				return 0;
 			}
 		}
 		unless (defined($redLen)) {
-			Warn "%s: %s", "$!", $src;
+			Warn "%s: %s", "$!", EscapedPath($src);
 			close $sfh;
 			close $dfh;
 			return 0;
@@ -1918,7 +2112,7 @@ sub CopyFile {
 		close $sfh;
 		unless (close $dfh) {
 			# マウントしている場合など close に失敗することがある
-			Warn "Can't close: %s", $dst;
+			Warn "Can't close: %s", EscapedPath($dst);
 			return 0;
 		}
 	}
@@ -1979,6 +2173,8 @@ unless (exists $ENV{HOME} and defined($ENV{HOME})) {
 	'maintain-scan-data'	=> 0,
 	# タイムアウト連続発生許容回数
 	'limit-timeout'	=> 3,
+	# リソースフォークをスキャンするかどうか
+	rsrc	=> 0,
 
 	#--- clamdscan オプション ----
 	# 冗長モード
@@ -2095,7 +2291,7 @@ EOF
 				$Conf{bufsiz} = 1024 * 1024 * $1;
 			}
 			else {
-				Error 'Value "%s" invalid for option %s', $name, $value;
+				Error 'Value "%s" invalid for option %s', $value, $name;
 				exit $ExitStatus{BadParams};
 			}
 		},
@@ -2113,6 +2309,18 @@ EOF
 		'maintain-scan-data'	=> \$Conf{'maintain-scan-data'},
 		# タイムアウト連続発生許容回数
 		'limit-timeout=i'	=> \$Conf{'limit-timeout'},
+		# リソースフォークスキャンをするかどうか
+		'scan-rsrc'	=> sub {
+			my	($name, $value) = @_;
+
+			if (defined $Sys::Filesystem::VERSION) {
+				$Conf{rsrc} = 1;
+			}
+			else {
+				Error 'Option "%s" requires Sys::Filesystem', $name;
+				exit $ExitStatus{UnsupportedOpts};
+			}
+		},
 	},
 
 	# clamdscan オプション
@@ -2345,6 +2553,12 @@ EOF
 
 	# timeout 連続発生回数
 	timeoutFreq	=> 0,
+
+	# ファイルシステム
+	filessystem	=> [],
+
+	# HFS フォーマットのファイルシステムのディレクトリ上かどうか
+	onHfsDir	=> 0,
 );
 
 #-------------------------------------------------------------------------------
@@ -2542,7 +2756,7 @@ B<ClamdOmitScan.pl> uses UNIX domain socket to connect to clamd.
 When this option is not specified, B<ClamdOmitScan.pl> uses C<LocalSocket>
 value, if it is written in F<clamd.conf>.
 
-DEFAULT: F</tmp/clamd>
+DEFAULT: F</tmp/clamd.socket>
 
 =item --port I<port> | --port=I<port>
 
@@ -2659,6 +2873,14 @@ normally.
 
 DEFAULT: 3
 
+=item --scan-rsrc
+
+Scan a resource fork of each file on HFS filesystem.
+
+This option requires perl module L<Sys::Filesystem|Sys::Filesystem>.
+
+DEFAULT: (no)
+
 =back
 
 =head2 CLAMDSCAN OPTIONS
@@ -2749,13 +2971,16 @@ C<ClamOmitScan.pl --data=/var/ClamdOmitScan/scanData --maintain-scan-data>
 
 42: Bad argument passed.
 
+43: Unsupported option passed.
+
 50: Error occured for scanning.
 
 51: Error occured for maintenance scan data.
 
 =head1 SEE ALSO
 
-L<clamd(8)>, L<clamdscan(1)>, L<clamscan(1)>, L<flock(2)>, L<clamd.conf(5)>
+L<clamd(8)>, L<clamdscan(1)>, L<clamscan(1)>, L<flock(2)>, L<clamd.conf(5)>,
+L<Sys::Filesystem|Sys::Filesystem>
 
 =head1 AUTHOR
 
